/* This file was generated by upb_generator from the input file:
 *
 *     google/protobuf/json_enumvalue_options.proto
 *
 * Do not edit -- your changes will be discarded when the file is
 * regenerated.
 * NO CHECKED-IN PROTOBUF GENCODE */

#include <stddef.h>
#include "upb/generated_code_support.h"
#include "google/protobuf/json_enumvalue_options.upb_minitable.h"
#include "google/protobuf/descriptor.upb_minitable.h"

// Must be last.
#include "upb/port/def.inc"

extern const UPB_PRIVATE(upb_GeneratedExtensionListEntry)* UPB_PRIVATE(upb_generated_extension_list);
typedef struct {
  upb_MiniTableField fields[1];
} pb__enumvalue__JsonEnumValueOptions_msg_init_Fields;

static const pb__enumvalue__JsonEnumValueOptions_msg_init_Fields pb_enumvalue_JsonEnumValueOptions__fields = {{
  {1, UPB_SIZE(12, 16), 64, kUpb_NoSub, 9, (int)kUpb_FieldMode_Scalar | ((int)kUpb_FieldRep_StringView << kUpb_FieldRep_Shift)},
}};

const upb_MiniTable pb__enumvalue__JsonEnumValueOptions_msg_init = {
  &pb_enumvalue_JsonEnumValueOptions__fields.fields[0],
  UPB_SIZE(24, 32), 1, kUpb_ExtMode_NonExtendable, 1, UPB_FASTTABLE_MASK(8), 0,
#ifdef UPB_TRACING_ENABLED
  "pb.enumvalue.JsonEnumValueOptions",
#endif
  UPB_FASTTABLE_INIT({
    {0x0000000000000000, &_upb_FastDecoder_DecodeUnknown},
    {0x001000000000000a, &upb_DecodeFast_String_Scalar_Tag1Byte},
  })
};

static const upb_MiniTableExtension pb_enumvalue_json_ext_obj = {
  {998, 0, 0, UPB_SIZE(3, 4), 11, (int)kUpb_FieldMode_Scalar | (int)kUpb_LabelFlags_IsExtension | ((int)UPB_SIZE(kUpb_FieldRep_4Byte, kUpb_FieldRep_8Byte) << kUpb_FieldRep_Shift)},
  {.UPB_PRIVATE(submsg) = &pb__enumvalue__JsonEnumValueOptions_msg_init},
  &google__protobuf__EnumValueOptions_msg_init,

};
UPB_LINKARR_APPEND(upb_AllExts)
const upb_MiniTableExtension* pb_enumvalue_json_ext = &pb_enumvalue_json_ext_obj;
  static const upb_MiniTable *messages_layout_google_protobuf_json_enumvalue_options_proto_upb_file_layout[1] = {
  &pb__enumvalue__JsonEnumValueOptions_msg_init,
};


static const upb_MiniTableExtension *extensions_layout_google_protobuf_json_enumvalue_options_proto_upb_file_layout[1] = {
  &pb_enumvalue_json_ext_obj,
};

UPB_LINKARR_DECLARE(upb_AllExts, const upb_MiniTableExtension*);
UPB_CONSTRUCTOR(upb_GeneratedRegistry_Constructor, google_protobuf_json_enumvalue_options_proto_ext_constructor, {
  static bool finished = false;
  if (finished) return;
  finished = true;
  static UPB_PRIVATE(upb_GeneratedExtensionListEntry) entry = {
    UPB_LINKARR_START(upb_AllExts),
    UPB_LINKARR_STOP(upb_AllExts),
    NULL
  };
  UPB_ASSERT(entry.next == NULL);
  entry.next = UPB_PRIVATE(upb_generated_extension_list);
  UPB_PRIVATE(upb_generated_extension_list) = &entry;
})
const upb_MiniTableFile google_protobuf_json_enumvalue_options_proto_upb_file_layout = {
  messages_layout_google_protobuf_json_enumvalue_options_proto_upb_file_layout,
  NULL,
  extensions_layout_google_protobuf_json_enumvalue_options_proto_upb_file_layout,
  1,
  0,
  1,
};

#include "upb/port/undef.inc"

